/*  
 * Copyright 2012 Andy Botting <andy@andybotting.com>
 *  
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *  
 * This file is distributed in the hope that it will be useful, but  
 * WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
 * General Public License for more details.  
 *  
 * You should have received a copy of the GNU General Public License  
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.  
 *  
 * This file incorporates work covered by the following copyright and  
 * permission notice:
 * 
 * Copyright 2010 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.andybotting.tramhunter.activity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;
import java.util.SortedMap;
import java.util.TreeMap;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.location.Criteria;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.AsyncTask;
import android.os.Bundle;
import android.util.Log;
import android.view.ContextMenu;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnCreateContextMenuListener;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemClickListener;
import android.widget.BaseAdapter;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;

import com.actionbarsherlock.app.ActionBar;
import com.actionbarsherlock.app.SherlockListActivity;
import com.actionbarsherlock.view.Menu;
import com.actionbarsherlock.view.MenuInflater;
import com.actionbarsherlock.view.MenuItem;
import com.andybotting.tramhunter.R;
import com.andybotting.tramhunter.TramHunterConstants;
import com.andybotting.tramhunter.dao.TramHunterDB;
import com.andybotting.tramhunter.location.LastLocationFinder;

import com.andybotting.tramhunter.objects.Stop;
import com.andybotting.tramhunter.objects.StopsList;

public class NearStopsActivity extends SherlockListActivity implements LocationListener {

	private static final String TAG = "NearStopsActivity";
	private static final boolean LOGV = Log.isLoggable(TAG, Log.INFO);

	private final static int CONTEXT_MENU_VIEW_STOP = 0;
	
	private final String ACTIVITY_TITLE = "Nearest Stops"; // ±

	private ListView mListView;
	private List<Stop> mAllStops;
	private StopsList mNearStopsList;
	private LocationManager mLocationManager;
	protected LocationListener mLocationListener;

	private TramHunterDB mDB;

	private boolean mShowBusy = true;

	protected Criteria mCriteria;
	protected LastLocationFinder lastLocationFinder;

	/**
	 * On Create
	 */
	public void onCreate(Bundle icicle) {
		super.onCreate(icicle);

		setContentView(R.layout.near_stops_list);

		// Set up the Action Bar
		ActionBar actionBar = getSupportActionBar();
		actionBar.setHomeButtonEnabled(true);
		actionBar.setDisplayHomeAsUpEnabled(true);
		actionBar.setTitle(ACTIVITY_TITLE);

		mListView = (ListView) this.findViewById(android.R.id.list);
		mListView.setOnItemClickListener(listView_OnItemClickListener);
		mListView.setOnCreateContextMenuListener(mListView_OnCreateContextMenuListener);

		mDB = new TramHunterDB();

		// Get our stops from the DB
		mAllStops = mDB.getAllStops();

		// Make our Near Stops List for the map
		mNearStopsList = new StopsList();

	    // Instantiate a LastLocationFinder class.
	    // This will be used to find the last known location when the application starts.
	    lastLocationFinder = new LastLocationFinder(this);
	    lastLocationFinder.setChangedLocationListener(oneShotLastLocationUpdateListener);
	    
		// Get the location
		mLocationManager = (LocationManager) getSystemService(LOCATION_SERVICE);

		mCriteria = new Criteria();
		mCriteria.setAccuracy(Criteria.ACCURACY_FINE);
		
		if (!hasLocationEnabled())
			buildAlertNoLocationServices();
	
		// onResume will be called and getLocationAndUpdateStops() and
		// startLocationListening() will be called.
	}
  
	/**
	 * One-off location listener that receives updates from the
	 * {@link LastLocationFinder}. This is triggered where the last known
	 * location is outside the bounds of our maximum distance and latency.
	 */
	protected LocationListener oneShotLastLocationUpdateListener = new LocationListener() {
		public void onLocationChanged(Location l) {
			updateStops(l);
		}
		public void onProviderDisabled(String provider) {}
		public void onStatusChanged(String provider, int status, Bundle extras) {}
		public void onProviderEnabled(String provider) {}
	};
	
	/**
	 * Check Android settings that we have enabled location providers
	 */
	protected boolean hasLocationEnabled() {
		String bestAvailableProvider = mLocationManager.getBestProvider(mCriteria, true);
		if (bestAvailableProvider == null)
			return false;
		return true;
	}

	/**
	 * On activity resume
	 */
	@Override
	protected void onResume() {
		super.onResume();
		// Get the last known location (and optionally request location updates)
		// and update the stops list.
		if (hasLocationEnabled()) {
			getLocationAndUpdateStops();
			startLocationListening();
		}
	}

	/**
	 * On activity pause
	 */
	@Override
	protected void onPause() {
		// Stop listening for location updates when the Activity is inactive.
		stopLocationListening();
		super.onPause();
	}

	/**
	 * On activity destroy
	 */
	@Override
	protected void onDestroy() {
		super.onDestroy();
		mDB.close();
	}

	/**
	 * Build an alert dialog
	 */
	private void buildAlertNoLocationServices() {
		final AlertDialog.Builder builder = new AlertDialog.Builder(this);
		builder.setMessage("You do not have GPS or Wireless network location services enabled.\n\nWould you like to enable them now?").setTitle("No Location Services")
				.setCancelable(false).setPositiveButton("Yes", new DialogInterface.OnClickListener() {
					public void onClick(final DialogInterface dialog, final int id) {
						dialog.dismiss();
						launchLocationSettings();
					}
				}).setNegativeButton("No", new DialogInterface.OnClickListener() {
					public void onClick(final DialogInterface dialog, final int id) {
						dialog.cancel();
						finish();
					}
				});

		AlertDialog alert = builder.create();
		alert.show();
	}

	/**
	 * Open up the location settings
	 */
	private void launchLocationSettings() {
		final Intent intent = new Intent(android.provider.Settings.ACTION_LOCATION_SOURCE_SETTINGS);
		startActivityForResult(intent, -1);
	}

	/**
	 * View a stop
	 */
	private void viewStop(Stop stop) {
		int tramTrackerId = stop.getTramTrackerID();

		Bundle bundle = new Bundle();
		bundle.putInt("tramTrackerId", tramTrackerId);
		Intent intent = new Intent(NearStopsActivity.this, StopDetailsActivity.class);
		intent.putExtras(bundle);

		startActivityForResult(intent, 1);
	}

	/**
	 * Options menu
	 */
	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		MenuInflater inflater = getSupportMenuInflater();
		inflater.inflate(R.menu.near_stops, menu);
		return super.onCreateOptionsMenu(menu);
	}

	/**
	 * Options item select
	 */
	@Override
	public boolean onOptionsItemSelected(final MenuItem item) {
		switch (item.getItemId()) {
			case R.id.menu_map:
				showMap();
				break;
			case android.R.id.home:
				finish();
		}

		return false;
	}
	
	/**
	 * Show the map view, passing this stop
	 */
	private void showMap() {

		// Detect Google Maps
		try {
			Class.forName("com.google.android.maps.MapActivity");
			// Map view
			Bundle bundle = new Bundle();
			bundle.putParcelable("stopslist", mNearStopsList);
			Intent intent = new Intent(NearStopsActivity.this, StopMapActivity.class);
			intent.putExtras(bundle);
			startActivityForResult(intent, -1);
		} catch (Exception e) {
			Toast.makeText(this, "Google Maps are not available", Toast.LENGTH_LONG).show();
		}
	}

	/**
	 * On click of menu item
	 */
	private OnItemClickListener listView_OnItemClickListener = new OnItemClickListener() {
		public void onItemClick(AdapterView<?> adapterView, View row, int position, long id) {
			StopsListAdapter stopsListAdapter = (StopsListAdapter) adapterView.getAdapter();
			viewStop(stopsListAdapter.getStops().get(position));
		}
	};

	/**
	 * Create context menu
	 */
	private OnCreateContextMenuListener mListView_OnCreateContextMenuListener = new OnCreateContextMenuListener() {
		public void onCreateContextMenu(ContextMenu menu, View v, ContextMenu.ContextMenuInfo menuInfo) {
			menu.add(0, CONTEXT_MENU_VIEW_STOP, 0, "View Stop");
		}
	};
	
	
	/**
	 * Find the last known location and update the stops list
	 */
	protected void getLocationAndUpdateStops() {
		
		// This isn't directly affecting the UI, so put it on a worker thread.
		AsyncTask<Stop, Void, ArrayList<Stop>> findLastLocationTask = new AsyncTask<Stop, Void, ArrayList<Stop>>() {
			
			Location location; 
			
			// Can use UI thread here
			protected void onPreExecute() {
				if (mShowBusy) {
					// Show the dialog window
					mListView.setVisibility(View.GONE);
					findViewById(android.R.id.empty).setVisibility(View.GONE);
					findViewById(R.id.loading).setVisibility(View.VISIBLE);
					mShowBusy = false;
				}
			}			

			@Override
			protected ArrayList<Stop> doInBackground(final Stop... params) {
				
				// Find the last known location, specifying a required accuracy
				// of within the min distance between updates and a required
				// latency of the minimum time required between updates.
				location = lastLocationFinder.getLastBestLocation(TramHunterConstants.MAX_DISTANCE, System.currentTimeMillis() - TramHunterConstants.MAX_TIME);
				
				ArrayList<Stop> sortedStops = new ArrayList<Stop>();
				SortedMap<Double, Stop> sortedStopList = new TreeMap<Double, Stop>();

				for (Stop stop : mAllStops) {
					double distance = location.distanceTo(stop.getLocation());
					sortedStopList.put(distance, stop);
				}

				// Build a sorted list, of MAXSTOPS stops
				for (Entry<Double, Stop> item : sortedStopList.entrySet()) {
					Stop stop = item.getValue();

					// Don't show terminus stops > 8000
					if (stop.getTramTrackerID() < 8000) {
						sortedStops.add(stop);
						mNearStopsList.add(stop);
					}

					if (sortedStops.size() >= TramHunterConstants.MAX_NEARBY_STOPS)
						break;
				}

				return sortedStops;
			}
			
			// Can use UI thread here
			protected void onPostExecute(final ArrayList<Stop> sortedStops) {
				StopsListAdapter stopsListAdapter;
				// Refresh the entire list
				stopsListAdapter = new StopsListAdapter(sortedStops, location);
				setListAdapter(stopsListAdapter);

				// If we've just been showing the loading screen
				if (mListView.getVisibility() == View.GONE) {
					mListView.setVisibility(View.VISIBLE);
					findViewById(R.id.loading).setVisibility(View.GONE);
				}
			}
		};
		
		findLastLocationTask.execute();
	}	

	
	
	/**
	 * Given a location, update the stops list
	 */
	protected void updateStops(Location l) {
		
		final Location location = l;
		
		// This isn't directly affecting the UI, so put it on a worker thread.
		AsyncTask<Stop, Void, ArrayList<Stop>> updateStopsTask = new AsyncTask<Stop, Void, ArrayList<Stop>>() {
			
			// Can use UI thread here
			protected void onPreExecute() {
				if (mShowBusy) {
					// Show the dialog window
					mListView.setVisibility(View.GONE);
					findViewById(android.R.id.empty).setVisibility(View.GONE);
					findViewById(R.id.loading).setVisibility(View.VISIBLE);
					mShowBusy = false;
				}
			}			

			@Override
			protected ArrayList<Stop> doInBackground(final Stop... params) {
				
				ArrayList<Stop> sortedStops = new ArrayList<Stop>();
				SortedMap<Double, Stop> sortedStopList = new TreeMap<Double, Stop>();

				for (Stop stop : mAllStops) {
					double distance = location.distanceTo(stop.getLocation());
					sortedStopList.put(distance, stop);
				}

				// Build a sorted list
				for (Entry<Double, Stop> item : sortedStopList.entrySet()) {
					Stop stop = item.getValue();

					// Don't show terminus stops > 8000
					if (stop.getTramTrackerID() < 8000) {
						sortedStops.add(stop);
						mNearStopsList.add(stop);
					}

					if (sortedStops.size() >= TramHunterConstants.MAX_NEARBY_STOPS)
						break;
				}

				return sortedStops;
			}
			
			// Can use UI thread here
			protected void onPostExecute(final ArrayList<Stop> sortedStops) {
				StopsListAdapter stopsListAdapter;
				// Refresh the entire list
				stopsListAdapter = new StopsListAdapter(sortedStops, location);
				setListAdapter(stopsListAdapter);

				// If we've just been showing the loading screen
				if (mListView.getVisibility() == View.GONE) {
					mListView.setVisibility(View.VISIBLE);
					findViewById(R.id.loading).setVisibility(View.GONE);
				}
			}
		};
		
		updateStopsTask.execute();
	}	
	
	/**
	 * On location changed
	 */
	public void onLocationChanged(Location location) {
		
		if (location.distanceTo(location) > TramHunterConstants.DEFAULT_RADIUS) {
			updateStops(location);
		}
		
		if (location.hasAccuracy()) {
			getSupportActionBar().setTitle(ACTIVITY_TITLE + " (±" + (int) location.getAccuracy() + "m)");
		} else {
			getSupportActionBar().setTitle(ACTIVITY_TITLE);
		}	    
	}
	
	/**
     * Stop listening for location events
     */
	private void stopLocationListening() {
		if (LOGV) Log.i(TAG, "Stopping location listening");
		if (mLocationManager != null)
			mLocationManager.removeUpdates(this);
	}
	
	/**
	 * Start listening for location events
	 */
	private void startLocationListening() {
		if (mLocationManager != null) {

			String bestAvailableProvider = mLocationManager.getBestProvider(mCriteria, true);
			if (LOGV) Log.i(TAG, "Start location listening with provider: " + bestAvailableProvider);

			if (bestAvailableProvider == null) {
				buildAlertNoLocationServices();
			} else {
				mLocationManager.requestLocationUpdates(bestAvailableProvider, 0, 0, this);
			}
		}
	}

	/**
     * onProviderDisabled unused
     */
	@Override
	public void onProviderDisabled(String provider) {}

	/**
	 * onProviderEnabled unused
	 */
	@Override
	public void onProviderEnabled(String provider) {}

	/**
	 * onStatusChanged unused
	 */
	@Override
	public void onStatusChanged(String provider, int status, Bundle extras) {}
	
	
	/**
	 * Adapter for Stops list
	 */
	private class StopsListAdapter extends BaseAdapter {

		private ArrayList<Stop> mStops;
		private Location mLocation;

		public StopsListAdapter(ArrayList<Stop> stops, Location location) {
			mStops = stops;
			mLocation = location;
		}

		public ArrayList<Stop> getStops() {
			return mStops;
		}

		public int getCount() {
			return mStops.size();
		}

		public Object getItem(int position) {
			return position;
		}

		public long getItemId(int position) {
			return position;
		}

		public View getView(int position, View convertView, ViewGroup parent) {
			View pv;
			if (convertView == null) {
				LayoutInflater inflater = getLayoutInflater();
				pv = inflater.inflate(R.layout.near_stops_list_row, parent, false);
			} else {
				pv = convertView;
			}

			Stop stop = (Stop) mStops.get(position);

			String stopName = stop.getPrimaryName();
			String stopDetails = stop.getStopDetailsLine();
			String stopDistance = stop.formatDistanceTo(mLocation);

			((TextView) pv.findViewById(R.id.stop_name)).setText(stopName);
			((TextView) pv.findViewById(R.id.stop_details)).setText(stopDetails);
			((TextView) pv.findViewById(R.id.stop_distance)).setText(stopDistance);
			((TextView) pv.findViewById(R.id.stop_routes)).setText(stop.getRoutesString());

			return pv;
		}
	}

}
